import logging
import os
import shutil
from os import path

from questionary import prompt


logger = logging.getLogger("cfgcaddy.utils")


def user_confirm(question, default=True):
    """Ask the user to confirm a choice

    Args:
        question (string): a string that is presented to the user.
        default (string): the answer if the user just hits <Enter>.
            It must be True, False or None (meaning
            an answer is required of the user).

    Returns "answer" return value is True for "yes" or False for "no".
    """
    return prompt(
        [{"type": "confirm", "name": "ok", "message": question, "default": default}]
    ).get("ok")


def create_dirs(dirs=None):
    """Creates all folders in dirs

    Args:
        dirs ([string]): A list of paths to be linked

    Returns:
        None: Does not return anything
    """
    if dirs:
        for dir_name in dirs:
            try:
                os.makedirs(dir_name)
            except OSError:
                logger.error("Unable to create directory: {}", dir_name)


def make_parent_dirs(file_path):
    try:
        os.makedirs(path.dirname(file_path))
    except OSError:  # Python >2.5
        pass


def create_links(links=None):
    """Create symlinks for each item in links

    Args:
        links([Link]): a list of paths to link
    Returns:
        None: Does not return anything
    """
    if links:
        for link in links:
            try:
                make_parent_dirs(link.dest)
                os.symlink(link.src, link.dest)
            except OSError as err:
                # If the file exists and isn't a link
                if err.errno == 17 and not os.path.islink(link.dest):
                    logger.error(
                        "Can't make link from {} to {} because {}".format(
                            link.src, link.dest, err.strerror
                        )
                    )


def move_files(transactions=None):
    """Move the files in the list from src to dest

    Args:
        transactions([Link])

    Returns:
        None: Does not return anything
    """
    if transactions:
        for transaction in transactions:
            try:
                shutil.move(transaction.src, transaction.dest)
            except OSError:
                logger.error(
                    "Can't move from {} to {}".format(transaction.src, transaction.dest)
                )


def expand_path(file_path):
    return os.path.expandvars(os.path.expanduser(file_path))



import logging
import os
import shutil
from os import path

import pathspec

import cfgcaddy
import cfgcaddy.utils as utils
from cfgcaddy.link import Link

logger = logging.getLogger("cfgcaddy.linker")


def find_absences(src, dest, ignored_patterns=None):
    """Walk the source directory and return a lists of files and dirs absent
        from the destination directory

    Args:
        src: The path to copy from (Default is the script's location)
        dest The path to copy to (Defaults to home directory)

    Returns:
        absent_files: a list of Links
        absent_dirs: a list of paths to directories
    """
    if not ignored_patterns:
        ignored_patterns = []
    absent_dirs = []
    absent_files = []

    default_ignore = ["!.*", cfgcaddy.DEFAULT_CONFIG_NAME, ".git"]

    ignored = pathspec.PathSpec.from_lines(
        "gitwildmatch", default_ignore + ignored_patterns
    )

    for root, dirs, files in os.walk(src, topdown=True):
        rel_path = path.relpath(root, src)
        if rel_path == ".":
            rel_path = ""

        # Remove ignored directories from the walk
        dirs[:] = [
            dir_name
            for dir_name in dirs
            if not ignored.match_file(path.join(root, dir_name))
        ]
        files[:] = [f for f in files if not ignored.match_file(path.join(root, f))]

        # Create list of dirs that dont exist
        for dir_name in dirs:
            pathname = path.join(dest, rel_path, dir_name)
            if not path.exists(pathname):
                if path.islink(pathname):
                    os.unlink(pathname)  # Fix Broken Links
                absent_dirs.append(pathname)

        # Create a list of files to be symlinked
        for f in files:
            pathname = path.join(dest, rel_path, f)
            if not path.exists(pathname):
                if path.islink(pathname):
                    os.unlink(pathname)  # Fix Broken Links
                # Add the source and destination for the symlink
                absent_files.append(Link(path.join(root, f), pathname))

    return absent_files, absent_dirs


def link_folder(src, dest, force=False):
    """Link the folder src to the destination dest

    Args:
        src (path) - The path to link from
        dest (path) - The path to link to

    Returns:
        True if an operation was performed, else False
    """
    folder_name = path.basename(src).strip(".")

    try:
        # Both Folders Exist
        if path.exists(dest) and path.exists(src) and not path.islink(dest):
            if force or utils.user_confirm("Link and merge {} to {}".format(src, dest)):
                absent_files, absent_dirs = find_absences(dest, src)
                zip_file = shutil.make_archive(
                    path.join(src, "{}_backup".format(folder_name)),
                    "zip",
                    root_dir=dest,
                )
                logger.info("Backing up {} to {}".format(folder_name, zip_file))
                utils.create_dirs(absent_dirs)
                logger.info("Created {}".format(absent_dirs))
                utils.move_files(absent_files)
                logger.info("Moving files {}".format(absent_files))
                shutil.rmtree(dest)
                logger.info("Removed {}".format(dest))
                os.symlink(src, dest)
                logger.info("Symlinked {} to {}".format(src, dest))
                return True

        # Only the source exists
        elif not path.exists(dest) and not path.islink(dest) and path.exists(src):
            try:
                os.makedirs(path.dirname(dest), exist_ok=True)
            except Exception as e:
                logger.error("Failed to make dir {} - {}".format(dest, e))
                return
            os.symlink(src, dest)
            logger.info("Symlinked {} to {}".format(src, dest))
            return True

        # Only the destination exists
        elif path.exists(dest) and not path.exists(src):
            if force or utils.user_confirm(
                "Delete, Move to {} and" " Link back to {}?".format(src, dest)
            ):
                shutil.move(dest, src)
                logger.info("Moving {} to {}".format(dest, src))
                os.symlink(src, dest)
                logger.info("Symlinked {} to {}".format(src, dest))
                return True

        # Nothing to do
        return False

    except OSError as e:
        logger.error("Failed to link {} to {} - {}".format(src, dest, e))
        return True


class Linker:
    """Tyler Stapler's Config Linker"""

    def __init__(self, linker_config, prompt=True):

        if not linker_config:
            raise Exception("Linker requires Config!")

        self.config = linker_config
        self.prompt = prompt

        self.custom_links = self.config.links
        self.ignored_patterns = self.config.ignore_patterns

    def create_links(self):
        """Symlinks configuration files to the destination directory

        Parses the ignore file for regexes and then generates a list of files
        which need to be simulinked"""

        # TODO: Rewrite find_absences
        absent_files, absent_dirs = find_absences(
            self.config.linker_src, self.config.linker_dest, self.ignored_patterns
        )

        if len(absent_files) == 0:
            logger.info("No files to move")
            return

        logger.info("Preparing to symlink the following files")
        print("\n".join(link.dest for link in absent_files))
        if not self.prompt or utils.user_confirm("Are these the correct files?"):
            utils.create_dirs(dirs=absent_dirs)
            utils.create_links(links=absent_files)

    def create_custom_links(self):
        """Link all the files in the customlink file

        Returns:
            None
        """
        modified = False

        for file in self.custom_links:
            if path.isdir(file.src):
                if link_folder(file.src, file.dest, self.prompt):
                    modified = True
            if path.isfile(file.src):
                if utils.create_links([file]):
                    modified = True

        if not modified:
            logger.info("No folders to link")



import logging
import os

from cfgcaddy import LinkMode
from cfgcaddy.utils import make_parent_dirs

logger = logging.getLogger("cfgcaddy.link")


class Link:
    def __init__(self, src, dest):
        """docstring for __init__"""

        self.src = src
        self.dest = dest

        if not os.path.exists(src):
            raise Exception("The source directory needs to exist!")

    def __repr__(self):
        return "{} => {}".format(self.src, self.dest)

    @property
    def is_linked(self):
        return os.path.islink(self.dest)

    def create(self, mode=LinkMode.SKIP):
        if os.path.exists(self.dest):
            pass
        # TODO: if dest exists
        # TODO: if dest is a link
        # TODO: If override
        # TODO: If skip

        # TODO: if dest is a file
        # TODO: If override
        # TODO: If skip

        # TODO: if dest is a folder

        # TODO: if src is a folder
        # TODO: If override
        # TODO: If skip
        # TODO: if src is a file
        # TODO: If override
        # TODO: If skip
        else:
            # if dest doesnt exist create it
            try:
                make_parent_dirs(self.dest)
                os.symlink(self.src, self.dest)
            except OSError as err:
                logger.error(
                    "Can't make link from {} to {} because {}".format(
                        self.src, self.dest, err.strerror
                    )
                )



import glob
import logging
import os
import platform
import sys
from os import path

from ruamel.yaml import YAML

from cfgcaddy import utils
from cfgcaddy.link import Link

logger = logging.getLogger("cfgcaddy.config")
yaml = YAML(typ="safe")  # default, if not specified, is 'rt' (round-trip)


MISSING_FILE_MESSAGE = (
    "Please create a {} file in the same" " directory as the linker python file."
)


class LinkerConfig:
    config = None
    links = []

    def __init__(self, config_file_path=None, default_config=None):
        self.config_file_path = config_file_path

        if default_config:
            self.config = default_config
        elif self.config_file_path:
            self.read_config()

        # Don't continue if we don't have a config
        if not self.config:
            logger.error("No config has been provided or loaded")
            sys.exit(1)

        if not self.linker_src or not self.linker_dest:
            logger.error("You need to specify a src and destination")
            sys.exit(1)

        self.generate_links(self.links_yaml)

    @property
    def preferences(self):
        return self.config.get("preferences") or {}

    @property
    def linker_src(self):
        src = "linker_src"
        return utils.expand_path(self.preferences.get(src))

    @property
    def linker_dest(self):
        dest = "linker_dest"
        return utils.expand_path(self.preferences.get(dest))

    def write_config(self, prompt=True):
        if not os.path.exists(self.config_file_path) or (
            not prompt
            or utils.user_confirm(
                "The file {} exists.\n"
                "Would you like to overwrite this file?".format(self.config_file_path)
            )
        ):
            try:
                logger.info("Writing config file")
                with open(self.config_file_path, "w") as file:
                    logger.debug(self.config_file_path)
                    yaml.dump(self.config, file)
            except Exception as e:
                logger.error("Error writing config: {}".format(e))

    def read_config(self):
        if os.path.isfile(self.config_file_path):
            with open(self.config_file_path, "r") as file:
                self.config = yaml.load(file)

    def generate_links(self, links):
        custom_links = []

        for link in links:
            try:
                if link.get("os") and platform.system() not in link.get("os"):
                    logger.debug("OS: {}".format(link.get("os")))
                    continue

                link_src = utils.expand_path(link["src"])
                link_destinations = link.get("dest")
                src_files = glob.glob(path.join(self.linker_src, link_src))

                # To account for no destination
                if not link_destinations:
                    if len(src_files) > 1:
                        link_destinations = [path.dirname(link_src)]
                    else:
                        link_destinations = [link_src]
                for src_path in src_files:
                    for dest in map(utils.expand_path, link_destinations):
                        if len(src_files) > 1:
                            src_name = path.join(dest, path.basename(src_path))
                        else:
                            src_name = dest
                        if path.isabs(dest):
                            dest_path = src_name
                        else:
                            dest_path = path.join(self.linker_dest, src_name)
                        custom_links.append(
                            Link(
                                utils.expand_path(src_path),
                                utils.expand_path(dest_path),
                            )
                        )
            except KeyError:
                logger.exception("Bad custom link")

        logger.debug("Custom Links => {}".format(custom_links))
        self.links = custom_links

    @property
    def links_yaml(self):
        """Parse the YAML config representation into a consistent format"""
        links = []
        for link in self.config.get("links", []):
            if type(link) is str:
                link = {"src": link, "dest": []}
            if type(link.get("dest")) is str:
                link["dest"] = [link["dest"]]
            links.append(link)
        logger.debug("Links before formatting: {}".format(self.config.get("links")))
        logger.debug("Links after formatting: {}".format(links))
        return links

    @property
    def ignore_patterns(self):
        lines = self.config.get("ignore", [])
        """Parse the gitignore style file
        """
        logger.debug("Ignore Patterns: {}".format(lines))
        return lines
